from typing import List, Union, TypedDict
import exposedfunctionality.function_parser.types as exf_types
import funcnodes as fn


class DataFrameDict(TypedDict):
    columns: list[str]
    index: List[Union[str, int, float]]
    data: List[List[Union[str, int, float]]]


exf_types.add_type("DataFrameDict", DataFrameDict)


class SepEnum(fn.DataEnum):
    COMMA = ","
    SEMICOLON = ";"
    TAB = "\t"
    SPACE = " "
    PIPE = "|"

    def __str__(self):
        return str(self.value)


class DecimalEnum(fn.DataEnum):
    COMMA = ","
    DOT = "."

    def __str__(self):
        return str(self.value)


exf_types.add_type("pd.SepEnum", SepEnum)
exf_types.add_type("pd.DecimalEnum", DecimalEnum)



import funcnodes as fn

# import funcnodes_numpy to register the types
import funcnodes_numpy as fnnp  # noqa: F401
import pandas as pd
from exposedfunctionality.function_parser.types import type_to_string
from .dataframe import (
    NODE_SHELF as DF_SHELF,
    to_dict,
    from_dict,
    from_csv_str,
    from_csv_auto,
    GetColumnNode as get_column,
    SetColumnNode as set_column,
    to_orient_dict,
    from_orient_dict,
    df_iloc,
    df_ilocs,
    get_rows,
    GetRowNode as df_loc,
    SetRowNode as set_row,
    to_csv_str,
    df_from_array,
    DfFromExcelNode,
    df_to_xls,
    dropna,
    ffill,
    bfill,
    fillna,
    drop_duplicates,
    numeric_only,
    drop_columns,
    drop_rows,
    reduce_df,
    DropColumnNode as drop_column,
    DropRowNode as drop_row,
    add_column,
    add_row,
    df_concatenate,
    df_merge,
    filter,
    mask,
    df_join,
    # rows and columns
    df_rename_col,
    df_rename_cols_valid_identifier,
    # end rows and columns
    # math
    df_corr,
    df_cov,
    df_mean,
    df_median,
    df_std,
    df_sum,
    df_var,
    df_quantile,
    df_describe,
    df_value_counts,
    df_eval,
    # end math
)

from .dataseries import (
    ser_to_dict,
    ser_values,
    ser_to_list,
    ser_loc,
    ser_iloc,
    ser_from_dict,
    ser_from_list,
    NODE_SHELF as SERIES_SHELF,
    ser_str_contains,
    ser_str_startswith,
    ser_str_endswith,
    ser_str_count,
    ser_str_len,
    ser_str_extract,
    ser_str_find,
    ser_str_findall,
    ser_str_get,
    ser_str_replace,
    ser_str_slice,
    ser_str_split,
    ser_str_strip,
    ser_str_zfill,
    ser_str_pad,
    ser_str_center,
    ser_str_ljust,
    ser_str_rjust,
    ser_str_wrap,
    ser_str_repeat,
    ser_str_upper,
    ser_str_lower,
    ser_str_title,
    ser_str_capitalize,
    ser_str_swapcase,
    ser_str_cat,
)

from .grouping import (
    GroupByColumnNode as group_by_column,
    group_by,
    gr_mean,
    gr_sum,
    gr_max,
    gr_min,
    gr_std,
    gr_var,
    gr_count,
    gr_describe,
    group_to_list,
    GetDFfromGroupNode as get_df_from_group,
    NODE_SHELF as GROUPING_SHELF,
)


def encode_pdDf(obj, preview=False):
    if isinstance(obj, pd.DataFrame):
        if preview:
            return fn.Encdata(
                obj.head().to_dict(orient="split"),
                handeled=True,
                continue_preview=False,
            )
        else:
            return fn.Encdata(
                obj.to_dict(orient="split"),
                handeled=True,
                continue_preview=False,
            )
    if isinstance(obj, pd.Series):
        return fn.Encdata(
            obj.values,
            handeled=True,
        )
    return fn.Encdata(obj, handeled=False)


fn.JSONEncoder.add_encoder(encode_pdDf, [pd.DataFrame, pd.Series])


NODE_SHELF = fn.Shelf(
    nodes=[],
    subshelves=[DF_SHELF, SERIES_SHELF, GROUPING_SHELF],
    name="Pandas",
    description="Pandas nodes",
)

FUNCNODES_RENDER_OPTIONS: fn.RenderOptions = {
    "typemap": {
        type_to_string(pd.DataFrame): "table",
        type_to_string(pd.Series): "list",
    },
}

__version__ = "0.2.10"

__all__ = [
    "NODE_SHELF",
    "to_dict",
    "from_dict",
    "from_csv_str",
    "from_csv_auto",
    "get_column",
    "to_orient_dict",
    "from_orient_dict",
    "df_iloc",
    "df_loc",
    "ser_to_dict",
    "ser_values",
    "ser_to_list",
    "ser_loc",
    "ser_iloc",
    "ser_from_dict",
    "ser_from_list",
    "SERIES_SHELF",
    "DF_SHELF",
    "to_csv_str",
    "df_from_array",
    "DfFromExcelNode",
    "df_to_xls",
    "dropna",
    "ffill",
    "bfill",
    "fillna",
    "drop_duplicates",
    "df_corr",
    "numeric_only",
    "drop_columns",
    "df_concatenate",
    "drop_rows",
    "drop_column",
    "drop_row",
    "reduce_df",
    "add_column",
    "add_row",
    "df_describe",
    "df_merge",
    "filter",
    "set_column",
    "set_row",
    "mask",
    "df_cov",
    "df_quantile",
    "df_value_counts",
    "df_eval",
    "df_corr",
    "df_join",
    "df_ilocs",
    "get_rows",
    # df
    # rows and columns
    "df_rename_col",
    "df_rename_cols_valid_identifier",
    # end rows and columns
    # math
    "df_corr",
    "df_cov",
    "df_mean",
    "df_median",
    "df_std",
    "df_sum",
    "df_var",
    "df_quantile",
    "df_describe",
    "df_value_counts",
    "df_eval",
    # end math
    # end df
    # series
    # str
    "ser_str_contains",
    "ser_str_startswith",
    "ser_str_endswith",
    "ser_str_count",
    "ser_str_len",
    "ser_str_extract",
    "ser_str_find",
    "ser_str_findall",
    "ser_str_get",
    "ser_str_replace",
    "ser_str_slice",
    "ser_str_split",
    "ser_str_strip",
    "ser_str_zfill",
    "ser_str_pad",
    "ser_str_center",
    "ser_str_ljust",
    "ser_str_rjust",
    "ser_str_wrap",
    "ser_str_repeat",
    "ser_str_upper",
    "ser_str_lower",
    "ser_str_title",
    "ser_str_capitalize",
    "ser_str_swapcase",
    "ser_str_cat",
    # end str
    # end series
    # grouping
    "group_by",
    "gr_mean",
    "gr_sum",
    "gr_max",
    "gr_min",
    "gr_std",
    "gr_var",
    "gr_count",
    "gr_describe",
    "group_to_list",
    "group_by_column",
    "get_df_from_group",
    # end grouping
]



import pandas as pd
import funcnodes as fn
from pandas.api.typing import DataFrameGroupBy
from typing import List


@fn.NodeDecorator(
    node_id="pd.gr.groupby_column",
    name="Group By Column",
    description="Groups a DataFrame by a column.",
    outputs=[{"name": "group"}],
    default_io_options={
        "df": {
            "on": {
                "after_set_value": fn.decorator.update_other_io(
                    "column",
                    lambda x: list(x.columns),
                )
            }
        },
    },
)
def GroupByColumnNode(
    df: pd.DataFrame,
    column: str,
) -> pd.Series:
    df = df.copy()
    return df.groupby(column)


@fn.NodeDecorator(
    node_id="pd.gr.groupby",
    name="Group By",
    description="Groups a DataFrame by a column.",
    outputs=[{"name": "grouped", "type": pd.DataFrame}],
)
def group_by(
    df: pd.DataFrame,
    by: str,
) -> DataFrameGroupBy:
    sep = [s.strip() for s in by.split(",")]
    df = df.copy()
    return df.groupby(sep)


@fn.NodeDecorator(
    node_id="pd.gr.mean",
    name="Mean",
    description="Calculates the mean of a DataFrameGroup.",
    outputs=[{"name": "mean", "type": pd.DataFrame}],
)
def gr_mean(
    group: DataFrameGroupBy,
) -> pd.DataFrame:
    return group.mean()


@fn.NodeDecorator(
    node_id="pd.gr.sum",
    name="Sum",
    description="Calculates the sum of a DataFrameGroup.",
    outputs=[{"name": "sum", "type": pd.DataFrame}],
)
def gr_sum(
    group: DataFrameGroupBy,
) -> pd.DataFrame:
    return group.sum()


@fn.NodeDecorator(
    node_id="pd.gr.max",
    name="Max",
    description="Calculates the max of a DataFrameGroup.",
    outputs=[{"name": "max", "type": pd.DataFrame}],
)
def gr_max(
    group: DataFrameGroupBy,
) -> pd.DataFrame:
    return group.max()


@fn.NodeDecorator(
    node_id="pd.gr.min",
    name="Min",
    description="Calculates the min of a DataFrameGroup.",
    outputs=[{"name": "min", "type": pd.DataFrame}],
)
def gr_min(
    group: DataFrameGroupBy,
) -> pd.DataFrame:
    return group.min()


@fn.NodeDecorator(
    node_id="pd.gr.std",
    name="Standard Deviation",
    description="Calculates the standard deviation of a DataFrameGroup.",
    outputs=[{"name": "std", "type": pd.DataFrame}],
)
def gr_std(
    group: DataFrameGroupBy,
) -> pd.DataFrame:
    return group.std()


@fn.NodeDecorator(
    node_id="pd.gr.var",
    name="Variance",
    description="Calculates the variance of a DataFrameGroup.",
    outputs=[{"name": "var", "type": pd.DataFrame}],
)
def gr_var(
    group: DataFrameGroupBy,
) -> pd.DataFrame:
    return group.var()


@fn.NodeDecorator(
    node_id="pd.gr.count",
    name="Count",
    description="Calculates the count of a DataFrameGroup.",
    outputs=[{"name": "count", "type": pd.DataFrame}],
)
def gr_count(
    group: DataFrameGroupBy,
) -> pd.DataFrame:
    return group.count()


@fn.NodeDecorator(
    node_id="pd.gr.describe",
    name="Describe",
    description="Describes a DataFrameGroup.",
    outputs=[
        {
            "name": "description",
        }
    ],
)
def gr_describe(
    group: DataFrameGroupBy,
) -> pd.DataFrame:
    return group.describe()


@fn.NodeDecorator(
    node_id="pd.gr.group_to_list",
    name="Group to List",
    description="Converts a DataFrameGroup to a list of DataFrames.",
    outputs=[{"name": "list", "type": list}],
)
def group_to_list(
    group: DataFrameGroupBy,
) -> List[pd.DataFrame]:
    return [group for _, group in group]


@fn.NodeDecorator(
    node_id="pd.gr.get_df_from_group",
    name="Get DataFrame from Group",
    description="Gets a DataFrame from a DataFrameGroup.",
    outputs=[{"name": "df"}],
    default_io_options={
        "group": {
            "on": {
                "after_set_value": fn.decorator.update_other_io(
                    "name",
                    lambda x: list(x.groups.keys()),
                )
            }
        },
    },
)
def GetDFfromGroupNode(
    group: DataFrameGroupBy,
    name: str,
) -> pd.DataFrame:
    df = group.get_group(name).copy()
    return df


NODE_SHELF = fn.Shelf(
    nodes=[
        GroupByColumnNode,
        group_by,
        gr_mean,
        gr_sum,
        gr_max,
        gr_min,
        gr_std,
        gr_var,
        gr_count,
        gr_describe,
        group_to_list,
        GetDFfromGroupNode,
    ],
    name="Grouping",
    description="Pandas grouping nodes",
    subshelves=[],
)



import funcnodes as fn


from ._convert import (  # noqa: F401
    to_dict,
    from_dict,
    from_csv_str,
    from_csv_auto,
    to_csv_str,
    to_orient_dict,
    from_orient_dict,
    df_from_array,
    DfFromExcelNode,
    df_to_xls,
    CONVERT_SHELF,
)
from ._manipulation import (  # noqa: F401
    dropna,
    fillna,
    bfill,
    ffill,
    drop_duplicates,
    numeric_only,
    DropColumnNode,
    DropRowNode,
    drop_columns,
    drop_rows,
    reduce_df,
    add_column,
    add_row,
    df_concatenate,
    df_merge,
    df_join,
    MANIPULATE_SHELF,
)
from ._math import (  # noqa: F401
    df_corr,
    df_cov,
    df_mean,
    df_median,
    df_std,
    df_sum,
    df_var,
    df_quantile,
    df_describe,
    df_value_counts,
    df_eval,
    MATH_SHELF,
)

from ._rows_cols import (  # noqa: F401
    GetColumnNode,
    SetColumnNode,
    GetRowNode,
    SetRowNode,
    df_iloc,
    get_rows,
    df_ilocs,
    df_rename_col,
    df_rename_cols_valid_identifier,
    ROW_COLS_SHELF,
)
from ._masking import (  # noqa: F401
    filter,
    mask,
    MASK_SHELF,
)

NODE_SHELF = fn.Shelf(
    nodes=[],
    name="Datataframe",
    description="Pandas DataFrame nodes",
    subshelves=[
        CONVERT_SHELF,
        MANIPULATE_SHELF,
        ROW_COLS_SHELF,
        MASK_SHELF,
        MATH_SHELF,
    ],
)



import pandas as pd
import funcnodes as fn
from typing import Any, List
from ..utils import to_valid_identifier

# region cols


@fn.NodeDecorator(
    node_id="pd.get_column",
    name="Get Column",
    description="Gets a column from a DataFrame.",
    outputs=[{"name": "series", "type": pd.Series}],
    default_io_options={
        "df": {
            "on": {
                "after_set_value": fn.decorator.update_other_io(
                    "column",
                    lambda x: list(iter(x)),
                )
            }
        },
    },
)
def GetColumnNode(df: pd.DataFrame, column: str) -> pd.Series:
    return df[column]


@fn.NodeDecorator(
    node_id="pd.set_column",
    name="Set Column",
    description="Sets a column in a DataFrame.",
    default_io_options={
        "df": {
            "on": {
                "after_set_value": fn.decorator.update_other_io(
                    "column",
                    lambda x: list(iter(x)),
                )
            }
        },
    },
)
def SetColumnNode(df: pd.DataFrame, column: str, data: Any) -> pd.DataFrame:
    df = df.copy()
    df[column] = data
    return df


# endregion cols


# region rows


@fn.NodeDecorator(
    node_id="pd.get_row",
    name="Get Row",
    description="Gets a row from a DataFrame by label.",
    outputs=[{"name": "series", "type": pd.Series}],
    default_io_options={
        "df": {
            "on": {
                "after_set_value": fn.decorator.update_other_io(
                    "row",
                    lambda x: list(x.index),
                )
            }
        },
    },
)
def GetRowNode(df: pd.DataFrame, row: str) -> pd.Series:
    return df.loc[df.index.to_list()[0].__class__(row)]  # transform to the correct type


@fn.NodeDecorator(
    node_id="pd.get_rows",
    name="Get Rows",
    description="Gets rows from a DataFrame by label.",
)
def get_rows(
    df: pd.DataFrame,
    rows: List[Any],
) -> pd.DataFrame:
    rows = [df.index.to_list()[0].__class__(row) for row in rows]
    return df.loc[rows]


@fn.NodeDecorator(
    node_id="pd.set_row",
    name="Set Row",
    description="Sets a row in a DataFrame.",
    default_io_options={
        "df": {
            "on": {
                "after_set_value": fn.decorator.update_other_io(
                    "row",
                    lambda x: list(x.index),
                )
            }
        },
    },
)
def SetRowNode(df: pd.DataFrame, row: str, data: Any) -> pd.DataFrame:
    df = df.copy()
    df.loc[df.index.to_list()[0].__class__(row)] = data
    return df


@fn.NodeDecorator(
    node_id="pd.df_iloc",
    name="Get Row by Index",
    description="Gets a row from a DataFrame by index.",
    outputs=[{"name": "row", "type": pd.Series}],
    default_io_options={
        "df": {
            "on": {
                "after_set_value": lambda src, result: src.node[
                    "index"
                ].update_value_options(min=0, max=len(result) - 1, step=1)
            }
        },
    },
)
def df_iloc(
    df: pd.DataFrame,
    index: int = 0,
) -> pd.Series:
    return df.iloc[int(index)]


@fn.NodeDecorator(
    node_id="pd.df_locs",
    name="Get Rows by Indices",
    description="Gets rows from a DataFrame by indices.",
    outputs=[{"name": "rows", "type": pd.DataFrame}],
)
def df_ilocs(
    df: pd.DataFrame,
    indices: List[int],
) -> pd.DataFrame:
    return df.iloc[[int(i) for i in indices]]


# endregion rows


@fn.NodeDecorator(
    node_id="pd.df_rename_col",
    name="Rename Column",
    description="Renames a column in a DataFrame.",
    default_io_options={
        "df": {
            "on": {
                "after_set_value": fn.decorator.update_other_io(
                    "old_name",
                    lambda x: x.columns.to_list(),
                )
            }
        },
    },
)
def df_rename_col(
    df: pd.DataFrame,
    old_name: str,
    new_name: str,
) -> pd.DataFrame:
    return df.rename(columns={old_name: new_name})


@fn.NodeDecorator(
    "pd.df_rename_cols_valid_identifier",
    name="Rename Columns to Valid Identifiers",
    description="Renames columns in a DataFrame to valid identifiers.",
)
def df_rename_cols_valid_identifier(
    df: pd.DataFrame,
) -> pd.DataFrame:
    return df.rename(
        columns={
            col: to_valid_identifier(
                col,
            )
            for col in df.columns
        }
    )


ROW_COLS_SHELF = fn.Shelf(
    nodes=[
        GetColumnNode,
        SetColumnNode,
        GetRowNode,
        SetRowNode,
        df_iloc,
        get_rows,
        df_ilocs,
        df_rename_col,
        df_rename_cols_valid_identifier,
    ],
    name="Rows and Columns",
    description="OPeration on rows and columns",
    subshelves=[],
)



import pandas as pd
import funcnodes as fn
from typing import List


@fn.NodeDecorator(
    node_id="pd.filter",
    name="Filter",
    description="Filters a DataFrame based on a condition.",
    outputs=[{"name": "filtered", "type": pd.DataFrame}],
)
def filter(
    df: pd.DataFrame,
    condition: str,
) -> pd.DataFrame:
    return df.query(condition)


@fn.NodeDecorator(
    node_id="pd.mask",
    name="Mask",
    description="Masks a DataFrame based on a condition.",
    outputs=[{"name": "masked", "type": pd.DataFrame}],
)
def mask(
    df: pd.DataFrame,
    mask: List[bool],
) -> pd.DataFrame:
    return df[mask]


MASK_SHELF = fn.Shelf(
    nodes=[
        filter,
        mask,
    ],
    name="Masking",
    description="Masking operations on DataFrames",
    subshelves=[],
)



from typing import List, Union, Dict, Any, Tuple
import chardet
import pandas as pd
from io import StringIO
import numpy as np


def detect_encoding(bytes):
    """Detect file encoding using chardet."""
    result = chardet.detect(bytes)
    return result["encoding"]


def get_lines(file: str):
    with open(file, "rb") as rawdata:
        data = rawdata.read()

    encoding = detect_encoding(data)

    with open(file, "r", encoding=encoding, errors="replace") as f:
        lines = f.readlines()

    return lines, encoding


def guess_table_info(
    line,
    possible_delimiters,
    possible_decimal_separators,
    possible_thousands_separators,
):
    lineinfo = {
        "length": len(line),
    }

    for delim in possible_delimiters:
        for dec_sep in possible_decimal_separators:
            if delim == dec_sep:
                continue
            for thou_sep in possible_thousands_separators:
                if delim == thou_sep:
                    continue
                if dec_sep == thou_sep:
                    continue

                # try to read the line to Series
                try:
                    series = pd.read_csv(
                        StringIO(line),
                        sep=delim,
                        decimal=dec_sep,
                        thousands=thou_sep,
                        header=None,
                        index_col=False,
                        engine="python",
                    )
                    lineinfo[(delim, dec_sep, thou_sep)] = series
                except Exception:
                    continue
    return lineinfo


def guess_best_table_params(
    tableinfos: List[dict], checklines: List[str], cutoff_ratio: float = 0.5
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    tableinfo_summary = {}
    number_of_lines = len(checklines)

    # fill tableinfo_summary with None values where there is no data
    for tableinfo in tableinfos:
        for parsedata, series in tableinfo.items():
            if not isinstance(series, pd.DataFrame):
                continue
            if parsedata not in tableinfo_summary:
                tableinfo_summary[parsedata] = [None] * number_of_lines

    for i, tableinfo in enumerate(tableinfos):
        for parsedata, series in tableinfo.items():
            if not isinstance(series, pd.DataFrame):
                continue
            tableinfo_summary[parsedata][i] = series

    # first we count the number of valid data reads for set of reding parameters
    parameter_lengths = {
        k: len([x for x in v if x is not None]) for k, v in tableinfo_summary.items()
    }
    max_length = max(parameter_lengths.values())
    cutoff = int(np.ceil(max_length * cutoff_ratio))
    # we drop all that are below the cutoff
    for paml, param_len in parameter_lengths.items():
        if param_len < cutoff:
            tableinfo_summary.pop(paml)

    # next we filter by the mean data length, mean to make sure not a single long line is weighted to much,
    # but readings that result in the simple case of a single entry should also removed compared to many longer ones
    series_lengths = {
        k: float(np.mean([len(x.columns) for x in v if x is not None]))
        for k, v in tableinfo_summary.items()
    }
    max_mean_ser_length = max(series_lengths.values()) * cutoff_ratio
    for paml, param_len in series_lengths.items():
        if param_len < max_mean_ser_length:
            tableinfo_summary.pop(paml)

    series_lengths = {
        k: [len(x.columns) for x in v if x is not None]
        for k, v in tableinfo_summary.items()
    }
    highest_series_length_count = {
        pname: max(serl, key=serl.count) for pname, serl in series_lengths.items()
    }

    # next we filter by type of data, where flaot is prefered over int and int over other
    total_types = {}
    for k, v in tableinfo_summary.items():
        _max_length = highest_series_length_count[k]
        relevant_series = [
            x for x in v if x is not None and len(x.columns) == _max_length
        ]
        dtypes = []
        for s in relevant_series:
            dtypes.append(list(s.dtypes))

        dtypes = np.array(dtypes).T.tolist()
        number_ints = [sum([1 for x in d if x == np.dtype("int64")]) for d in dtypes]
        number_floats = [
            sum([1 for x in d if x == np.dtype("float64")]) for d in dtypes
        ]
        number_other = [
            sum([1 for x in d if x not in [np.dtype("int64"), np.dtype("float64")]])
            for d in dtypes
        ]

        total_types[k] = {
            "int": sum(number_ints),
            "float": sum(number_floats),
            "other": sum(number_other),
            "numerical": sum(number_ints) + sum(number_floats),
        }

    higest_numerical_value = max([v["numerical"] for v in total_types.values()])
    for k, v in list(total_types.items()):
        if v["numerical"] < higest_numerical_value:
            total_types.pop(k)

    highest_float_value = max([v["float"] for v in total_types.values()])
    count_cutff = int(np.ceil(highest_float_value * cutoff_ratio))
    for k, v in list(total_types.items()):
        if v["float"] < count_cutff:
            total_types.pop(k)

    highest_int_value = max([v["int"] for v in total_types.values()])
    count_cutff = int(np.ceil(highest_int_value * cutoff_ratio))
    for k, v in list(total_types.items()):
        if v["int"] < count_cutff:
            total_types.pop(k)

    highest_other_value = max([v["other"] for v in total_types.values()])
    count_cutff = int(np.ceil(highest_other_value * cutoff_ratio))
    for k, v in list(total_types.items()):
        if v["other"] < count_cutff:
            total_types.pop(k)

    # next we filter by the number of parameters that fullyfil the above criteria and take the longest one
    parameter_lengths = {
        k: len([x for x in v if x is not None]) for k, v in tableinfo_summary.items()
    }
    max_length = max(parameter_lengths.values())
    for paml, param_len in parameter_lengths.items():
        if param_len < max_length:
            total_types.pop(paml)

    parser_args = list(total_types.keys())[0]
    parser_args

    # now we check where the last line of header data is, that has other lengths than the data of the table
    last_drop_line = -1
    exp_series = highest_series_length_count[parser_args]

    for i, sers in enumerate(tableinfo_summary[parser_args]):
        if sers is None:
            last_drop_line = i
        else:
            if len(sers.columns) != exp_series:
                last_drop_line = i

    # lastly we estimate the header, if we expect numerical values, all initial rows without numerical values are header
    remaining_lines = checklines[last_drop_line + 1 :]
    headerlines = -1
    for line in remaining_lines:
        _df = pd.read_csv(
            StringIO(line),
            sep=parser_args[0],
            decimal=parser_args[1],
            thousands=parser_args[2],
            header=None,
            engine="python",
        )
        typeslist = _df.dtypes.to_list()

        if highest_float_value > 0:
            if np.dtype("float64") in typeslist:
                break
            else:
                headerlines += 1
                continue
        elif highest_int_value > 0:
            if np.dtype("int64") in typeslist:
                break
            else:
                headerlines += 1
                continue
        break

    if headerlines == -1:
        headerlines = None
    headerlines

    # return the estimated parameters
    return {
        "sep": parser_args[0],
        "decimal": parser_args[1],
        "thousands": parser_args[2],
        "header": headerlines,
        "skiprows": last_drop_line + 1,
    }, {
        "exp_series": exp_series,
        "last_drop_line": last_drop_line,
        "headerlines": headerlines,
    }


def auto_parse_table(
    source: Union[str, bytes],
    possible_delimiters: List[str] = None,
    possible_decimal_separators: List[str] = None,
    possible_thousands_separators: List[str] = None,
    max_lines: int = 200,
    cutoff_ratio: float = 0.5,
) -> Tuple[pd.DataFrame, Dict[str, Any]]:
    encoding = None
    if isinstance(source, bytes):
        encoding = detect_encoding(source)
        stringdata = source.decode(encoding).strip()
        lines = stringdata.split("\n")
    elif isinstance(source, str):
        source = source.strip()
        lines = source.split("\n")
        stringdata = source
    else:
        raise ValueError("source must be either a string or bytes")

    if not possible_delimiters:
        possible_delimiters = [
            ",",
            "\t",
            "\\t",
            " " * 4,
            " " * 3,
            " " * 2,
            " " * 1,
            ";",
            "|",
        ]

    if not possible_decimal_separators:
        possible_decimal_separators = [".", ",", None]
    if not possible_thousands_separators:
        possible_thousands_separators = [None, ",", ".", " "]

    if max_lines > len(lines):
        max_lines = len(lines) - 1

    checklines = lines[:max_lines]
    tableinfos = [
        guess_table_info(
            line=line,
            possible_delimiters=possible_delimiters,
            possible_decimal_separators=possible_decimal_separators,
            possible_thousands_separators=possible_thousands_separators,
        )
        for line in checklines
    ]

    parse_params, auto_params = guess_best_table_params(
        tableinfos=tableinfos,
        checklines=checklines,
        cutoff_ratio=cutoff_ratio,
    )
    if encoding:
        parse_params["encoding"] = encoding

    try:
        df = pd.read_csv(StringIO(stringdata), **parse_params)
    except Exception as e:
        raise ValueError(
            f"Could not parse table: {e} with params: {parse_params},{auto_params}"
        )
    return df, parse_params



import pandas as pd
import funcnodes as fn
from typing import Optional, Literal, Union, Any, List
import numpy as np
# region nan


@fn.NodeDecorator(
    node_id="pd.dropna",
    name="Drop NA",
    description="Drops rows or columns with NA values.",
)
def dropna(
    df: pd.DataFrame,
    axis: Literal["index", "columns"] = "index",
    how: Literal["any", "all"] = "any",
    subset: Optional[str] = None,
) -> pd.DataFrame:
    if subset is not None:
        subset = [s.strip() for s in subset.split(",")]

    return df.dropna(axis=axis, how=how, subset=subset)


@fn.NodeDecorator(
    node_id="pd.fillna",
    name="Fill NA",
    description="Fills NA values with a specified value.",
)
def fillna(
    df: pd.DataFrame,
    value: Union[str, int, float] = 0,
) -> pd.DataFrame:
    return df.fillna(value)


@fn.NodeDecorator(
    node_id="pd.bfill",
    name="Backfill",
    description="Backfills NA values.",
)
def bfill(
    df: pd.DataFrame,
) -> pd.DataFrame:
    return df.bfill()


@fn.NodeDecorator(
    node_id="pd.ffill",
    name="Forwardfill",
    description="Forwardfills NA values.",
)
def ffill(
    df: pd.DataFrame,
) -> pd.DataFrame:
    return df.ffill()


# endregion nan


# region duplicates


@fn.NodeDecorator(
    node_id="pd.drop_duplicates",
    name="Drop Duplicates",
    description="Drops duplicate rows.",
)
def drop_duplicates(
    df: pd.DataFrame,
    subset: Optional[str] = None,
) -> pd.DataFrame:
    if subset is not None:
        subset = [s.strip() for s in subset.split(",")]
    return df.drop_duplicates(subset=subset)


# endregion duplicates


# region filter


@fn.NodeDecorator(
    node_id="pd.numeric_only",
    name="Numeric Only",
)
def numeric_only(df: pd.DataFrame, label_encode: bool = False) -> pd.DataFrame:
    """
    Converts a DataFrame to only hold numeric values.
    Optionally, non-numeric values can be converted to numeric labels.

    Parameters:
    - df: pandas DataFrame
    - label_encode: bool, if True, convert non-numeric values to numeric labels

    Returns:
    - A new DataFrame containing only numeric values
    """

    df = df.copy()
    for column in df.select_dtypes(exclude=[np.number]):
        try:
            df[column] = pd.to_numeric(df[column])
        except ValueError:
            pass

    if label_encode:
        for column in df.select_dtypes(include=["object", "category"]):
            df[column] = df[column].astype("category").cat.codes

    numeric_df = df.select_dtypes(include=[np.number])
    return numeric_df


# endregion filter

# region drop


@fn.NodeDecorator(
    node_id="pd.drop_column",
    name="Drop Column",
    description="Drops a column from a DataFrame.",
    default_io_options={
        "df": {
            "on": {
                "after_set_value": fn.decorator.update_other_io(
                    "column",
                    lambda x: list(x.columns),
                )
            }
        },
    },
)
def DropColumnNode(df: pd.DataFrame, column: str) -> pd.DataFrame:
    return df.drop(column, axis=1)


@fn.NodeDecorator(
    node_id="pd.drop_row",
    name="Drop Row",
    description="Drops a row from a DataFrame.",
    default_io_options={
        "df": {
            "on": {
                "after_set_value": fn.decorator.update_other_io(
                    "row",
                    lambda x: list(x.index),
                )
            }
        },
    },
)
def DropRowNode(df: pd.DataFrame, row: str) -> pd.DataFrame:
    return df.drop(df.index.to_list()[0].__class__(row), axis=0)


@fn.NodeDecorator(
    node_id="pd.drop_columns",
    name="Drop Columns",
    description="Drops columns from a DataFrame.",
)
def drop_columns(
    df: pd.DataFrame,
    columns: str,
) -> pd.DataFrame:
    columns = [s.strip() for s in columns.split(",")]
    return df.drop(columns, axis=1)


@fn.NodeDecorator(
    node_id="pd.drop_rows",
    name="Drop Rows",
    description="Drops rows from a DataFrame.",
)
def drop_rows(
    df: pd.DataFrame,
    rows: str,
) -> pd.DataFrame:
    rows = [s.strip() for s in rows.split(",")]

    if len(df.index) == 0:
        return df
    cls = df.index.to_list()[0].__class__
    rows = [cls(row) for row in rows]

    return df.drop(rows, axis=0)


@fn.NodeDecorator(
    node_id="pd.reduce_df",
    name="Reduce DataFrame",
    description="Reduces the DataFrame by keeping rows where the specified columns change significantly.",
    outputs=[{"name": "reduced df", "type": pd.DataFrame}],
)
def reduce_df(
    df: pd.DataFrame,
    on: Union[str, List[str]],
    threshold: Optional[Union[float, List[float]]] = None,
    percentage_threshold: float = 0.01,
) -> pd.DataFrame:
    """
    Reduces the DataFrame by keeping rows where the specified columns change significantly.

    Parameters:
    df (pd.DataFrame): The input DataFrame.
    on (Union[str, List[str]]): Column(s) to monitor for significant changes.
                                If a single string is provided, it will be split by commas and whitespace stripped.
    threshold (Optional[Union[float, List[float]]]): Threshold for significant change.
                                Can be a single value or a list of values corresponding to `on`.
                                If None, the threshold is set to percentage_threshold of the min-max
                                range of each column.
    percentage_threshold (float): Percentage of the min-max range of each column to use as the threshold.
                                  Ignored if `threshold` is provided.

    Returns:
    pd.DataFrame: The reduced DataFrame.
    """
    # Handle case where `on` is a comma-separated string
    if isinstance(on, str):
        on = [c.strip() for c in on.split(",")]

    # Ensure `on` is a list of strings
    if not isinstance(on, list):
        on = [on]

    # Check for empty DataFrame or insufficient rows
    if df.empty:
        raise ValueError("Input DataFrame is empty.")
    if len(df) == 1:
        return df

    # Ensure the columns specified in `on` are present and numeric
    for col in on:
        if col not in df.columns:
            raise ValueError(f"Column '{col}' not found in DataFrame.")
        if not np.issubdtype(df[col].dtype, np.number):
            raise ValueError(f"Column '{col}' must contain numeric data.")

    # Calculate default thresholds if not provided
    if threshold is None:
        ranges = np.array([df[col].max() - df[col].min() for col in on])
        # Prevent thresholds from being zero by using a small epsilon value
        threshold = percentage_threshold * np.where(
            ranges == 0, np.finfo(float).eps, ranges
        )

    # Ensure threshold is the correct length
    if not isinstance(threshold, (list, np.ndarray)):
        threshold = [threshold] * len(on)
    threshold = np.array(threshold)

    if len(threshold) != len(on):
        raise ValueError(
            "Threshold must be a single value or a list of values equal in length to 'on'."
        )

    # Vectorize difference calculation for faster processing

    values = df[on].values

    current_row = 0
    keep = [current_row]
    while True:
        # use values[0] since values will be reassign to reduce the aize each iteration
        maxval = values[0] + threshold
        minval = values[0] - threshold

        row_out_ranges = (values > maxval).any(axis=1) | (values < minval).any(axis=1)
        if row_out_ranges.any():
            # first out of range index
            first_out_row = row_out_ranges.argmax()

            if first_out_row > 1:
                # if it is larger than 1, we keep the previous row to capture the change e.g. if it is a step
                first_out_row -= 1
            elif first_out_row == 0:
                # if the first row out is 0 the first elemnte is true,
                # which should not happen since we compare against it
                raise ValueError(
                    "The first row is out of range, this should not happen"
                )

            # set the current row to the first out of range row
            current_row += first_out_row
            # reduce the values to the remaining rows
            values = values[first_out_row:]
            # add the current row to the keep list
            keep.append(current_row)
        else:
            # if no row is out of range, we can break
            break

    # Ensure the last row is included
    if len(df) - 1 not in keep:
        keep.append(len(df) - 1)

    return df.iloc[keep].reset_index(drop=True)


# endregion drop

# region add


@fn.NodeDecorator(
    node_id="pd.add_column",
    name="Add Column",
    description="Adds a column from a DataFrame.",
)
def add_column(
    df: pd.DataFrame,
    column: str,
    data: Any,
) -> pd.DataFrame:
    df = df.copy()
    df[column] = data
    return df


@fn.NodeDecorator(
    node_id="pd.add_row",
    name="Add Row",
    description="Adds a row to a DataFrame.",
)
def add_row(
    df: pd.DataFrame,
    row: Union[dict, list],
) -> pd.DataFrame:
    if not isinstance(row, dict):
        try:
            row = {c: row[c] for c in df.columns}
        except Exception:
            pass
        if len(row) != len(df.columns):
            raise ValueError(
                "Row must have the same number of columns as the DataFrame"
            )
        row = {c: [v] for c, v in zip(df.columns, row)}
    df = pd.concat([df, pd.DataFrame(row)])
    return df


# endregion add


# region merge
@fn.NodeDecorator(
    node_id="pd.concat",
    name="Concatenate",
    description="Concatenates two DataFrames.",
)
def df_concatenate(df1: pd.DataFrame, df2: pd.DataFrame) -> pd.DataFrame:
    return pd.concat([df1, df2])


@fn.NodeDecorator(
    node_id="pd.merge",
    name="Merge",
    description="Merges two DataFrames.",
    outputs=[{"name": "df", "type": pd.DataFrame}],
    default_io_options={
        "df_left": {
            "on": {
                "after_set_value": fn.decorator.update_other_io(
                    "left_on",
                    lambda x: list(x.columns),
                )
            }
        },
        "df_right": {
            "on": {
                "after_set_value": fn.decorator.update_other_io(
                    "right_on",
                    lambda x: list(x.columns),
                )
            }
        },
    },
)
def df_merge(
    df_left: pd.DataFrame,
    df_right: pd.DataFrame,
    how: Literal["inner", "outer", "left", "right"] = "inner",
    left_on: Optional[str] = None,
    right_on: Optional[str] = None,
) -> pd.DataFrame:
    return pd.merge(df_left, df_right, how=how, left_on=left_on, right_on=right_on)


@fn.NodeDecorator(
    node_id="pd.join",
    name="Join",
    description="Joins two DataFrames.",
    outputs=[{"name": "df", "type": pd.DataFrame}],
)
def df_join(
    df_left: pd.DataFrame,
    df_right: pd.DataFrame,
    how: Literal["inner", "outer", "left", "right"] = "left",
    on: Optional[str] = None,
    lsuffix: str = "",
    rsuffix: str = "",
) -> pd.DataFrame:
    return df_left.join(df_right, how=how, on=on, lsuffix=lsuffix, rsuffix=rsuffix)


# endregion merge


MANIPULATE_SHELF = fn.Shelf(
    nodes=[
        dropna,
        fillna,
        bfill,
        ffill,
        drop_duplicates,
        numeric_only,
        DropColumnNode,
        DropRowNode,
        drop_columns,
        drop_rows,
        reduce_df,
        add_column,
        add_row,
        df_concatenate,
        df_merge,
        df_join,
    ],
    name="Manipulation",
    description="DataFrame manipulations",
    subshelves=[],
)



from typing import Union, Any, Optional
import funcnodes as fn
import numpy as np
import pandas
from ._str import SERIES_STR_SHELF
from ._str import *  # noqa: F401, F403


@fn.NodeDecorator(
    node_id="pd.ser_to_dict",
    name="To Dictionary",
    description="Converts a Series to a dictionary.",
    outputs=[{"name": "dict", "type": dict}],
)
def ser_to_dict(
    ser: pandas.Series,
) -> dict:
    return ser.to_dict()


@fn.NodeDecorator(
    node_id="pd.ser_values",
    name="Get Values",
    description="Gets the values of a Series.",
    outputs=[{"name": "values", "type": np.ndarray}],
)
def ser_values(
    ser: pandas.Series,
) -> np.ndarray:
    return ser.to_numpy(copy=True)


@fn.NodeDecorator(
    node_id="pd.ser_to_list",
    name="To List",
    description="Converts a Series to a list.",
    outputs=[{"name": "list", "type": list}],
)
def ser_to_list(
    ser: pandas.Series,
) -> list:
    return ser.to_list()


@fn.NodeDecorator(
    node_id="pd.ser_loc",
    name="Get Value",
    description="Gets a value from a Series by label.",
    outputs=[{"name": "value", "type": Any}],
    default_io_options={
        "ser": {
            "on": {
                "after_set_value": fn.decorator.update_other_io(
                    "label",
                    lambda x: list(x.index),
                )
            }
        },
    },
)
def ser_loc(
    ser: pandas.Series,
    label: str,
) -> str:
    # taransform label to the correct type
    label = ser.index.to_list()[0].__class__(label)
    return ser.loc[label]


@fn.NodeDecorator(
    node_id="pd.ser_iloc",
    name="Get Value by Index",
    description="Gets a value from a Series by index.",
    outputs=[{"name": "value", "type": Any}],
    default_io_options={
        "ser": {
            "on": {
                "after_set_value": lambda src, result: src.node[
                    "index"
                ].update_value_options(min=0, max=len(result) - 1, step=1)
            }
        },
    },
)
def ser_iloc(
    ser: pandas.Series,
    index: int,
) -> Union[str]:
    return ser.iloc[int(index)]


@fn.NodeDecorator(
    node_id="pd.ser_from_dict",
    name="From Dictionary",
    description="Creates a Series from a dictionary.",
    outputs=[{"name": "series", "type": pandas.Series}],
)
def ser_from_dict(
    data: dict,
    name: Optional[str] = None,
) -> pandas.Series:
    return pandas.Series(data, name=name)


@fn.NodeDecorator(
    node_id="pd.ser_from_list",
    name="From List",
    description="Creates a Series from a list.",
    outputs=[{"name": "series", "type": pandas.Series}],
)
def ser_from_list(
    data: list,
    name: Optional[str] = None,
) -> pandas.Series:
    return pandas.Series(data, name=name)


NODE_SHELF = fn.Shelf(
    name="Series",
    nodes=[
        ser_to_dict,
        ser_values,
        ser_to_list,
        ser_loc,
        ser_iloc,
        ser_from_dict,
        ser_from_list,
    ],
    description="Pandas Series nodes",
    subshelves=[SERIES_STR_SHELF],
)



import re


def to_valid_identifier(
    s,
    max_length=30,
    digit_prefix="_",
    drop_characters="!\"#$%&'()*+,./:;<=>?@[\\]^`{|}~",
    merge_underscores=True,
):
    # remove unwanted characters
    s = s.translate(str.maketrans("", "", drop_characters))

    # Remove any characters that are not alphanumeric or underscores
    s = re.sub(r"\W|^(?=\d)", "_", s)

    # Ensure the identifier doesn't start with a digit
    if s[0].isdigit():
        s = digit_prefix + s

    # Merge consecutive underscores
    if merge_underscores:
        s = re.sub("_+", "_", s)

    # Truncate the identifier to the maximum length
    s = s[:max_length]

    return s



import pandas as pd
import funcnodes as fn
from typing import Literal, List, Optional, Union

from ..utils import to_valid_identifier


@fn.NodeDecorator(
    node_id="pd.corr",
    name="Correlation",
    description="Calculates the correlation between columns.",
    outputs=[{"name": "correlation", "type": pd.DataFrame}],
)
def df_corr(
    df: pd.DataFrame,
    method: Literal["pearson", "kendall", "spearman"] = "pearson",
    numeric_only: bool = False,
) -> pd.DataFrame:
    return df.corr(method=method, numeric_only=numeric_only)


@fn.NodeDecorator(
    node_id="pd.cov",
    name="Covariance",
    description="Calculates the covariance between columns.",
    outputs=[{"name": "covariance", "type": pd.DataFrame}],
)
def df_cov(
    df: pd.DataFrame,
    min_periods: int = 1,
    ddof: int = 1,
    numeric_only: bool = False,
) -> pd.DataFrame:
    return df.cov(min_periods=min_periods, ddof=ddof, numeric_only=numeric_only)


@fn.NodeDecorator(
    node_id="pd.mean",
    name="Mean",
    description="Calculates the mean of entries.",
    outputs=[{"name": "mean", "type": pd.Series}],
)
def df_mean(
    df: pd.DataFrame,
    axis: Literal[None, 0, 1] = 0,
    numeric_only: bool = False,
) -> pd.Series:
    return df.mean(axis=axis, numeric_only=numeric_only)


@fn.NodeDecorator(
    node_id="pd.median",
    name="Median",
    description="Calculates the median of entries.",
    outputs=[{"name": "median", "type": pd.Series}],
)
def df_median(
    df: pd.DataFrame,
    axis: Literal[None, 0, 1] = 0,
    numeric_only: bool = False,
) -> pd.Series:
    return df.median(axis=axis, numeric_only=numeric_only)


@fn.NodeDecorator(
    node_id="pd.std",
    name="Standard Deviation",
    description="Calculates the standard deviation of entries.",
    outputs=[{"name": "std", "type": pd.Series}],
)
def df_std(
    df: pd.DataFrame,
    axis: Literal[None, 0, 1] = 0,
    numeric_only: bool = False,
) -> pd.Series:
    return df.std(axis=axis, numeric_only=numeric_only)


@fn.NodeDecorator(
    node_id="pd.sum",
    name="Sum",
    description="Calculates the sum of entries.",
    outputs=[{"name": "sum", "type": pd.Series}],
)
def df_sum(
    df: pd.DataFrame,
    axis: Literal[None, 0, 1] = 0,
    numeric_only: bool = False,
) -> pd.Series:
    return df.sum(axis=axis, numeric_only=numeric_only)


@fn.NodeDecorator(
    node_id="pd.var",
    name="Variance",
    description="Calculates the variance of entries.",
    outputs=[{"name": "var", "type": pd.Series}],
)
def df_var(
    df: pd.DataFrame,
    axis: Literal[None, 0, 1] = 0,
    numeric_only: bool = False,
) -> pd.Series:
    return df.var(axis=axis, numeric_only=numeric_only)


@fn.NodeDecorator(
    node_id="pd.quantile",
    name="Quantile",
    description="Calculates the quantile of entries.",
    outputs=[{"name": "quantile", "type": pd.Series}],
)
def df_quantile(
    df: pd.DataFrame,
    q: float = 0.5,
    axis: Literal[None, 0, 1] = 0,
    numeric_only: bool = False,
) -> pd.Series:
    return df.quantile(q=q, axis=axis, numeric_only=numeric_only)


@fn.NodeDecorator(
    node_id="pd.describe",
    name="Describe",
    description="Describes the DataFrame.",
    outputs=[{"name": "description", "type": pd.DataFrame}],
)
def df_describe(
    df: pd.DataFrame,
    percentiles: Optional[List[float]] = None,
) -> pd.DataFrame:
    return df.describe(
        percentiles=percentiles,
    )


@fn.NodeDecorator(
    node_id="pd.value_counts",
    name="Value Counts",
    description="Counts the occurrences of unique values.",
    outputs=[{"name": "value_counts", "type": pd.Series}],
)
def df_value_counts(
    df: pd.DataFrame,
    subset: Optional[str] = None,
    normalize: bool = False,
    sort: bool = True,
    ascending: bool = False,
    dropna: bool = True,
) -> pd.DataFrame:
    if subset is not None:
        subset = [s.strip() for s in subset.split(",")]
    multilabel_series = df.value_counts(
        subset=subset,
        normalize=normalize,
        sort=sort,
        ascending=ascending,
        dropna=dropna,
    )
    # convert to a dataframe with the indices as columns and the count column named "count"

    out_df = pd.DataFrame(multilabel_series).reset_index()
    out_df.columns = out_df.columns.astype(str)
    return out_df


@fn.NodeDecorator(
    node_id="pd.eval",
    name="Eval",
    description="Evaluates an expression in the context of the DataFrame.",
    outputs=[{"name": "result", "type": pd.Series}],
)
def df_eval(
    df: pd.DataFrame,
    expr: str,
) -> Union[pd.Series, pd.DataFrame]:
    locals = df.to_dict(orient="series")
    renamed_locals = df.rename(
        columns={col: to_valid_identifier(col) for col in df.columns}
    ).to_dict(orient="series")
    return pd.eval(
        expr,
        target=df,
        local_dict={**locals, **renamed_locals, "df": df},
        global_dict={},
    )


MATH_SHELF = fn.Shelf(
    nodes=[
        df_corr,
        df_cov,
        df_mean,
        df_median,
        df_std,
        df_sum,
        df_var,
        df_quantile,
        df_describe,
        df_value_counts,
        df_eval,
    ],
    name="Math",
    description="Math on DataFrames",
    subshelves=[],
)



import pandas as pd
import funcnodes as fn
from typing import Optional, Literal, Union, List, Tuple
from ._types import DataFrameDict, SepEnum, DecimalEnum
from io import StringIO, BytesIO
import numpy as np
from funcnodes_basic.strings import POSSIBLE_DECODINGS_TYPE
from ._autoreader import auto_parse_table
# region dict


@fn.NodeDecorator(
    node_id="pd.df_to_dict",
    name="To Dictionary",
    description="Converts a DataFrame to a dictionary.",
    outputs=[{"name": "dict", "type": DataFrameDict}],
)
def to_dict(
    df: pd.DataFrame,
) -> dict:
    return df.to_dict(orient="split")


@fn.NodeDecorator(
    node_id="pd.df_to_orient_dict",
    name="To Dictionary with Orientation",
    description="Converts a DataFrame to a dictionary with a specific orientation.",
    outputs=[{"name": "dict", "type": dict}],
)
def to_orient_dict(
    df: pd.DataFrame,
    orient: Literal["dict", "list", "split", "tight", "records", "index"] = "split",
) -> dict:
    return df.to_dict(orient=orient)


@fn.NodeDecorator(
    node_id="pd.df_from_dict",
    name="From Dictionary",
    description="Converts a dictionary to a DataFrame.",
    outputs=[{"name": "df", "type": pd.DataFrame}],
)
def from_dict(
    data: dict,
) -> pd.DataFrame:
    # from "split" orientation or from "thight" orientation
    if "columns" in data and "index" in data and "data" in data:
        df = pd.DataFrame(
            data["data"],
            columns=data["columns"],
            index=data["index"],
        )
        idxnames = data.get("index_names")
        if idxnames is not None and len(idxnames) == len(df.index):
            df.index.names = idxnames
        colnames = data.get("column_names")
        if colnames is not None and len(colnames) == len(df.columns):
            df.columns.names = colnames
        return df

    # by default we cannot distringuise between "dict" and "index" orientation since both have the same structure of
    # {column: {index: value}} or {index: {column: value}}
    # a small heuristic is to check if the first key is a string or not to determine the orientation
    if isinstance(data, list):
        return pd.DataFrame(data)
    if len(data) == 0:
        return pd.DataFrame()
    if isinstance(next(iter(data)), str):
        return pd.DataFrame(data)
    else:
        return pd.DataFrame(data).T


@fn.NodeDecorator(
    node_id="pd.df_from_orient_dict",
    name="From Dictionary with Orientation",
    description="Converts a dictionary with a specific orientation to a DataFrame.",
    outputs=[{"name": "df", "type": pd.DataFrame}],
)
def from_orient_dict(
    data: dict,
    orient: Literal["dict", "list", "split", "tight", "records", "index"] = "split",
) -> pd.DataFrame:
    if orient == "split":
        return pd.DataFrame(
            data.get("data"), columns=data.get("columns"), index=data.get("index")
        )
    elif orient in ["dict", "list", "records"]:
        return pd.DataFrame(data)
    elif orient == "tight":
        df = pd.DataFrame(
            data.get("data"), columns=data.get("columns"), index=data.get("index")
        )
        df.columns.names = data.get("column_names")
        df.index.names = data.get("index_names")
        return df
    elif orient == "index":
        return pd.DataFrame(data).T
    return pd.DataFrame(data)


# endregion dict


# region csv


@fn.NodeDecorator(
    node_id="pd.df_from_csv_str",
    name="From CSV",
    description="Reads a CSV file into a DataFrame.",
    outputs=[{"name": "df", "type": pd.DataFrame}],
    default_io_options={
        "skiprows": {
            "hidden": True,
        },
        "thousands": {
            "hidden": True,
        },
        "encoding": {
            "hidden": True,
        },
    },
)
def from_csv_str(
    source: Union[str, bytes],
    sep: SepEnum = ",",
    decimal: DecimalEnum = ".",
    thousands: Optional[DecimalEnum] = None,
    skiprows: Optional[int] = None,
    encoding: POSSIBLE_DECODINGS_TYPE = "utf-8",
) -> pd.DataFrame:
    sep = SepEnum.v(sep)
    decimal = DecimalEnum.v(decimal)
    thousands = DecimalEnum.v(thousands) if thousands is not None else None
    if isinstance(source, bytes):
        data = BytesIO(source)
    else:
        data = StringIO(source)
    return pd.read_csv(
        data,
        sep=sep,
        decimal=decimal,
        thousands=thousands,
        skiprows=skiprows,
        encoding=encoding,
    )


@fn.NodeDecorator(
    node_id="pd.df_to_csv_str",
    name="To CSV",
    description="Writes a DataFrame to a CSV string.",
    outputs=[{"name": "csv"}],
)
def to_csv_str(
    df: pd.DataFrame,
    sep: SepEnum = ",",
    decimal: DecimalEnum = ".",
    thousands: Optional[DecimalEnum] = None,
    index: bool = False,
) -> str:
    sep = SepEnum.v(sep)
    decimal = DecimalEnum.v(decimal)
    thousands = DecimalEnum.v(thousands) if thousands is not None else None

    return df.to_csv(sep=sep, decimal=decimal, index=index)


@fn.NodeDecorator(
    node_id="pd.df_from_csv_auto",
    name="From CSV Auto",
    description="Reads a CSV file into a DataFrame. Automatically detects the parameters.",
    outputs=[{"name": "df"}, {"name": "params"}],
)
def from_csv_auto(
    source: Union[str, bytes],
    possible_delimiters: List[str] = None,
    possible_decimal_separators: List[str] = None,
    possible_thousands_separators: List[str] = None,
    max_lines: int = 200,
    cutoff_ratio: float = 0.5,
) -> Tuple[pd.DataFrame, dict]:
    df, params = auto_parse_table(
        source,
        possible_delimiters=possible_delimiters,
        possible_decimal_separators=possible_decimal_separators,
        possible_thousands_separators=possible_thousands_separators,
        max_lines=max_lines,
        cutoff_ratio=cutoff_ratio,
    )
    return df, params


# endregion csv

# region excel


@fn.NodeDecorator(
    node_id="pd.df_from_xlsx",
    name="From Excel",
    description="Reads an Excel file into a DataFrame.",
    outputs=[{"name": "df", "type": pd.DataFrame}],
    default_io_options={
        "data": {
            "on": {
                "after_set_value": fn.decorator.update_other_io(
                    "sheet",
                    lambda x: pd.ExcelFile(BytesIO(x)).sheet_names,
                )
            }
        }
    },
)
def DfFromExcelNode(data: bytes, sheet: Optional[str] = None, with_index: bool = False):
    # if sheed is not provided, we return the first sheet
    if sheet is None:
        sheet = 0
    return pd.read_excel(
        BytesIO(data), sheet_name=sheet, index_col=0 if with_index else None
    )


@fn.NodeDecorator(
    node_id="pd.df_to_xls",
    name="To Excel",
    description="Writes a DataFrame to an Excel file.",
    outputs=[{"name": "xls"}],
)
def df_to_xls(
    df: pd.DataFrame,
    sheet_name: str = "Sheet1",
    with_index: bool = False,
    exceldata: Optional[BytesIO] = None,
) -> bytes:
    if exceldata is None:
        output = BytesIO()
        mode = "w"
    else:
        output = exceldata
        mode = "a"
    with pd.ExcelWriter(output, engine="openpyxl", mode=mode) as writer:
        df.to_excel(writer, sheet_name=sheet_name, index=with_index)
    return output.getvalue()


# endregion excel


# region array


@fn.NodeDecorator(
    node_id="pd.df_from_array",
    name="From Array",
    description="Creates a DataFrame from an array.",
    outputs=[{"name": "df", "type": pd.DataFrame}],
)
def df_from_array(
    data: Union[list[list[Union[str, int, float]]], np.ndarray],
    columns: List[str] = None,
    index: List[Union[str, int, float]] = None,
) -> pd.DataFrame:
    if columns is None:
        columns = [f"Col {i+1}" for i in range(len(data[0]))]
    return pd.DataFrame(data, columns=columns, index=index)


# endregion array


CONVERT_SHELF = fn.Shelf(
    nodes=[
        to_dict,
        from_dict,
        from_csv_auto,
        from_csv_str,
        to_csv_str,
        to_orient_dict,
        from_orient_dict,
        df_from_array,
        DfFromExcelNode,
        df_to_xls,
    ],
    name="Convert",
    description="Conversions for DataFrames",
    subshelves=[],
)



"""
string operations for pandas Series
"""

import pandas as pd
import funcnodes as fn
from typing import Any, List, Optional, Literal

# region filter


@fn.NodeDecorator(
    node_id="pd.str_contains",
    name="Contains",
    description="Checks if a string is contained in each element of a Series.",
)
def ser_str_contains(series: pd.Series, pat: str) -> pd.Series:
    return series.str.contains(pat)


@fn.NodeDecorator(
    node_id="pd.str_startswith",
    name="Starts With",
    description="Checks if each element of a Series starts with a string.",
)
def ser_str_startswith(series: pd.Series, pat: str) -> pd.Series:
    return series.str.startswith(pat)


@fn.NodeDecorator(
    node_id="pd.str_endswith",
    name="Ends With",
    description="Checks if each element of a Series ends with a string.",
)
def ser_str_endswith(series: pd.Series, pat: str) -> pd.Series:
    return series.str.endswith(pat)


# endregion filter

# region stats


@fn.NodeDecorator(
    node_id="pd.str_count",
    name="Count",
    description="Counts occurrences of a pattern in each element of a Series.",
)
def ser_str_count(series: pd.Series, pat: str) -> pd.Series:
    return series.str.count(pat)


@fn.NodeDecorator(
    node_id="pd.str_len",
    name="Length",
    description="Calculates the length of each element of a Series.",
)
def ser_str_len(series: pd.Series) -> pd.Series:
    return series.str.len()


# endregion stats


# region substring


@fn.NodeDecorator(
    node_id="pd.str_extract",
    name="Extract",
    description="Extracts a pattern from each element of a Series.",
)
def ser_str_extract(series: pd.Series, pat: str) -> pd.DataFrame:
    return series.str.extract(pat, expand=True)


@fn.NodeDecorator(
    node_id="pd.str_find",
    name="Find",
    description="Finds the first occurrence of a pattern in each element of a Series.",
)
def ser_str_find(series: pd.Series, pat: str) -> pd.Series:
    return series.str.find(pat)


@fn.NodeDecorator(
    node_id="pd.str_findall",
    name="Find All",
    description="Finds all occurrences of a pattern in each element of a Series.",
)
def ser_str_findall(series: pd.Series, pat: str) -> pd.Series:
    return series.str.findall(pat)


@fn.NodeDecorator(
    node_id="pd.str_get",
    name="Get",
    description="Gets the nth character of each element of a Series.",
)
def ser_str_get(series: pd.Series, index: int) -> pd.Series:
    return series.str.get(index)


# endregion substring

# region transform


@fn.NodeDecorator(
    node_id="pd.str_replace",
    name="Replace",
    description="Replaces a pattern in each element of a Series.",
)
def ser_str_replace(series: pd.Series, pat: str, repl: str) -> pd.Series:
    return series.str.replace(pat, repl)


@fn.NodeDecorator(
    node_id="pd.str_slice",
    name="Slice",
    description="Slices each element of a Series.",
)
def ser_str_slice(series: pd.Series, start: int, stop: int, step: int = 1) -> pd.Series:
    return series.str.slice(start, stop, step)


@fn.NodeDecorator(
    node_id="pd.str_split",
    name="Split",
    description="Splits each element of a Series.",
)
def ser_str_split(series: pd.Series, pat: str) -> pd.Series:
    return series.str.split(pat)


@fn.NodeDecorator(
    node_id="pd.str_strip",
    name="Strip",
    description="Strips whitespace from each element of a Series.",
)
def ser_str_strip(series: pd.Series) -> pd.Series:
    return series.str.strip()


@fn.NodeDecorator(
    node_id="pd.str_zfill",
    name="Zero Fill",
    description="Fills each element of a Series with zeros.",
)
def ser_str_zfill(series: pd.Series, width: int) -> pd.Series:
    return series.str.zfill(width)


@fn.NodeDecorator(
    node_id="pd.str_pad",
    name="Pad",
    description="Pads each element of a Series.",
)
def ser_str_pad(
    series: pd.Series,
    width: int,
    side: Literal["left", "right", "both"] = "left",
    fillchar: str = " ",
) -> pd.Series:
    return series.str.pad(width, side, fillchar)


@fn.NodeDecorator(
    node_id="pd.str_center",
    name="Center",
    description="Centers each element of a Series.",
)
def ser_str_center(series: pd.Series, width: int, fillchar: str = " ") -> pd.Series:
    return series.str.center(width, fillchar)


@fn.NodeDecorator(
    node_id="pd.str_ljust",
    name="Left Justify",
    description="Left justifies each element of a Series.",
)
def ser_str_ljust(series: pd.Series, width: int, fillchar: str = " ") -> pd.Series:
    return series.str.ljust(width, fillchar)


@fn.NodeDecorator(
    node_id="pd.str_rjust",
    name="Right Justify",
    description="Right justifies each element of a Series.",
)
def ser_str_rjust(series: pd.Series, width: int, fillchar: str = " ") -> pd.Series:
    return series.str.rjust(width, fillchar)


@fn.NodeDecorator(
    node_id="pd.str_wrap",
    name="Wrap",
    description="Wraps each element of a Series.",
)
def ser_str_wrap(series: pd.Series, width: int) -> pd.Series:
    return series.str.wrap(width)


@fn.NodeDecorator(
    node_id="pd.str_repeat",
    name="Repeat",
    description="Repeats each element of a Series.",
)
def ser_str_repeat(series: pd.Series, repeats: int) -> pd.Series:
    return series.str.repeat(repeats)


@fn.NodeDecorator(
    node_id="pd.str_upper",
    name="Upper",
    description="Converts each element of a Series to uppercase.",
)
def ser_str_upper(series: pd.Series) -> pd.Series:
    return series.str.upper()


@fn.NodeDecorator(
    node_id="pd.str_lower",
    name="Lower",
    description="Converts each element of a Series to lowercase.",
)
def ser_str_lower(series: pd.Series) -> pd.Series:
    return series.str.lower()


@fn.NodeDecorator(
    node_id="pd.str_title",
    name="Title",
    description="Converts each element of a Series to title case.",
)
def ser_str_title(series: pd.Series) -> pd.Series:
    return series.str.title()


@fn.NodeDecorator(
    node_id="pd.str_capitalize",
    name="Capitalize",
    description="Capitalizes each element of a Series.",
)
def ser_str_capitalize(series: pd.Series) -> pd.Series:
    return series.str.capitalize()


@fn.NodeDecorator(
    node_id="pd.str_swapcase",
    name="Swap Case",
    description="Swaps the case of each element of a Series.",
)
def ser_str_swapcase(series: pd.Series) -> pd.Series:
    return series.str.swapcase()


@fn.NodeDecorator(
    node_id="pd.str_cat",
    name="Concatenate",
    description="Concatenates strings in a Series.",
    default_io_options={
        "series": {
            "on": {
                "after_set_value": fn.decorator.update_other_io(
                    "others",
                    lambda x: list(iter(x)),
                )
            }
        },
    },
)
def ser_str_cat(
    series: pd.Series, sep: Optional[str] = None, others: Optional[List[Any]] = None
) -> str:
    return series.str.cat(
        others,
        sep=sep,
    )


# endregion transform


SERIES_STR_SHELF = fn.Shelf(
    name="String Operations",
    description="String operations for pandas Series.",
    nodes=[
        ser_str_contains,
        ser_str_startswith,
        ser_str_endswith,
        ser_str_count,
        ser_str_len,
        ser_str_extract,
        ser_str_find,
        ser_str_findall,
        ser_str_get,
        ser_str_replace,
        ser_str_slice,
        ser_str_split,
        ser_str_strip,
        ser_str_zfill,
        ser_str_pad,
        ser_str_center,
        ser_str_ljust,
        ser_str_rjust,
        ser_str_wrap,
        ser_str_repeat,
        ser_str_upper,
        ser_str_lower,
        ser_str_title,
        ser_str_capitalize,
        ser_str_swapcase,
        ser_str_cat,
    ],
    subshelves=[],
)
